r"""Datasets included with PySaRe

Datasets
--------
- ``WeibullUniformParameters`` -- A simulated datases with Weibull distributed
  event times  
"""

from .Dataset import Dataset
from . import datasets

