from typing import Optional, Union, Iterable, Callable, Literal, TypeVar, NewType, Any


Batch = NewType('Batch', Any)
AxisLike = NewType('AxisLike', Any)

X = NewType('X', Any)
T = NewType('T', Any)
E = NewType('E', Any)